# Copyright 2020-2024 Intel Corporation
# SPDX-License-Identifier: Apache-2.0


"""UnbalancedFederatedDataset module."""

from abc import abstractmethod
from typing import List

import numpy as np
from tqdm import trange

from openfl.utilities.data_splitters.data_splitter import DataSplitter


def get_label_count(labels, label):
    """Count the number of samples with a specific label in a labels array.

    Args:
        labels (np.ndarray): Array of labels.
        label (int or str): The label to count.

    Returns:
        int: The count of the label in the labels array.
    """
    return len(np.nonzero(labels == label)[0])


def one_hot(labels, classes):
    """Apply One-Hot encoding to labels.

    Args:
        labels (np.ndarray): Array of labels.
        classes (int): The total number of classes.

    Returns:
        np.ndarray: The one-hot encoded labels.
    """
    return np.eye(classes)[labels]


class NumPyDataSplitter(DataSplitter):
    """Base class for splitting numpy arrays of data.

    This class should be subclassed when creating specific data splitter
    classes.
    """

    @abstractmethod
    def split(self, data: np.ndarray, num_collaborators: int) -> List[List[int]]:
        """Split the data."""
        raise NotImplementedError


class EqualNumPyDataSplitter(NumPyDataSplitter):
    """Class for splitting numpy arrays of data evenly.

    Args:
        shuffle (bool, optional): Flag determining whether to shuffle the
            dataset before splitting. Defaults to True.
        seed (int, optional): Random numbers generator seed. Defaults to 0.
    """

    def __init__(self, shuffle=True, seed=0):
        """Initialize.

        Args:
            shuffle (bool): Flag determining whether to shuffle the dataset
                before splitting. Defaults to True.
            seed (int): Random numbers generator seed. Defaults to 0.
                For different splits on envoys, try setting different values
                for this parameter on each shard descriptor.
        """
        self.shuffle = shuffle
        self.seed = seed

    def split(self, data, num_collaborators):
        """Split the data."""
        np.random.seed(self.seed)
        idx = range(len(data))
        if self.shuffle:
            idx = np.random.permutation(idx)
        slices = np.array_split(idx, num_collaborators)
        return slices


class RandomNumPyDataSplitter(NumPyDataSplitter):
    """Class for splitting numpy arrays of data randomly.

    Args:
        shuffle (bool, optional): Flag determining whether to shuffle the
            dataset before splitting. Defaults to True.
        seed (int, optional): Random numbers generator seed. Defaults to 0.
    """

    def __init__(self, shuffle=True, seed=0):
        """Initialize.

        Args:
            shuffle (bool): Flag determining whether to shuffle the dataset
                before splitting. Defaults to True.
            seed (int): Random numbers generator seed. Defaults to 0.
                For different splits on envoys, try setting different values
                for this parameter on each shard descriptor.
        """
        self.shuffle = shuffle
        self.seed = seed

    def split(self, data, num_collaborators):
        """Split the data."""
        np.random.seed(self.seed)
        idx = range(len(data))
        if self.shuffle:
            idx = np.random.permutation(idx)
        random_idx = np.sort(np.random.choice(len(data), num_collaborators - 1, replace=False))

        return np.split(idx, random_idx)


class LogNormalNumPyDataSplitter(NumPyDataSplitter):
    """Class for splitting numpy arrays of data according to a LogNormal
    distribution.

    Unbalanced (LogNormal) dataset split.
    This split assumes only several classes are assigned to each collaborator.
    Firstly, it assigns classes_per_col * min_samples_per_class items of
    dataset to each collaborator so all of collaborators will have some data
    after the split.
    Then, it generates positive integer numbers by log-normal (power) law.
    These numbers correspond to numbers of dataset items picked each time from
    dataset and assigned to a collaborator.
    Generation is repeated for each class assigned to a collaborator.
    This is a parametrized version of non-i.i.d. data split in FedProx
    algorithm.
    Origin source: https://github.com/litian96/FedProx/blob/master/data/mnist/generate_niid.py#L30

    Args:
        mu (float): Distribution hyperparameter.
        sigma (float): Distribution hyperparameter.
        num_classes (int): Number of classes.
        classes_per_col (int): Number of classes assigned to each collaborator.
        min_samples_per_class (int): Minimum number of collaborator samples of
            each class.
        seed (int, optional): Random numbers generator seed. Defaults to 0.

    .. note::
        This split always drops out some part of the dataset!
        Non-deterministic behavior selects only random subpart of class items.
    """

    def __init__(
        self,
        mu,
        sigma,
        num_classes,
        classes_per_col,
        min_samples_per_class,
        seed=0,
    ):
        """Initialize the generator.

        Args:
            mu (float): Distribution hyperparameter.
            sigma (float): Distribution hyperparameter.
            classes_per_col (int): Number of classes assigned to each
                collaborator.
            min_samples_per_class (int): Minimum number of collaborator
                samples of each class.
            seed (int): Random numbers generator seed. Defaults to 0.
                For different splits on envoys, try setting different values
                for this parameter on each shard descriptor.
        """
        self.mu = mu
        self.sigma = sigma
        self.num_classes = num_classes
        self.classes_per_col = classes_per_col
        self.min_samples_per_class = min_samples_per_class
        self.seed = seed

    def split(self, data, num_collaborators):
        """Split the data.

        Args:
            data (np.ndarray): numpy-like label array.
            num_collaborators (int): number of collaborators to split data
                across.
                Should be divisible by number of classes in ``data``.
        """
        np.random.seed(self.seed)
        idx = [[] for _ in range(num_collaborators)]
        samples_per_col = self.classes_per_col * self.min_samples_per_class
        for col in range(num_collaborators):
            for c in range(self.classes_per_col):
                label = (col + c) % self.num_classes
                label_idx = np.nonzero(data == label)[0]
                slice_start = col // self.num_classes * samples_per_col
                slice_start += self.min_samples_per_class * c
                slice_end = slice_start + self.min_samples_per_class
                print(f"Assigning {slice_start}:{slice_end} of class {label} to {col} col...")
                idx[col] += list(label_idx[slice_start:slice_end])
        if any(len(i) != samples_per_col for i in idx):
            raise SystemError(
                f"""All collaborators should have {samples_per_col} elements
but distribution is {[len(i) for i in idx]}"""
            )

        props_shape = (
            self.num_classes,
            num_collaborators // self.num_classes,
            self.classes_per_col,
        )
        props = np.random.lognormal(self.mu, self.sigma, props_shape)
        num_samples_per_class = [
            [[get_label_count(data, label) - self.min_samples_per_class]]
            for label in range(self.num_classes)
        ]
        num_samples_per_class = np.array(num_samples_per_class)
        props = num_samples_per_class * props / np.sum(props, (1, 2), keepdims=True)
        for col in trange(num_collaborators):
            for j in range(self.classes_per_col):
                label = (col + j) % self.num_classes
                num_samples = int(props[label, col // self.num_classes, j])

                print(f"Trying to append {num_samples} samples of {label} class to {col} col...")
                slice_start = np.count_nonzero(data[np.hstack(idx)] == label)
                slice_end = slice_start + num_samples
                label_count = get_label_count(data, label)
                if slice_end < label_count:
                    label_subset = np.nonzero(data == (col + j) % self.num_classes)[0]
                    idx_to_append = label_subset[slice_start:slice_end]
                    idx[col] = np.append(idx[col], idx_to_append)
                else:
                    print(
                        f"Index {slice_end} is out of bounds "
                        f"of array of length {label_count}. Skipping..."
                    )
        print(f"Split result: {[len(i) for i in idx]}.")
        return idx


class DirichletNumPyDataSplitter(NumPyDataSplitter):
    """Class for splitting numpy arrays of data according to a Dirichlet
    distribution.

    Generates the random sample of integer numbers from dirichlet distribution
    until minimum subset length exceeds the specified threshold.
    This behavior is a parametrized version of non-i.i.d. split in FedMA
    algorithm.
    Origin source: https://github.com/IBM/FedMA/blob/master/utils.py#L96

    Args:
        alpha (float, optional): Dirichlet distribution parameter. Defaults
            to 0.5.
        min_samples_per_col (int, optional): Minimal amount of samples per
            collaborator. Defaults to 10.
        seed (int, optional): Random numbers generator seed. Defaults to 0.
    """

    def __init__(self, alpha=0.5, min_samples_per_col=10, seed=0):
        """Initialize.

        Args:
            alpha (float): Dirichlet distribution parameter. Defaults to 0.5.
            min_samples_per_col (int): Minimal amount of samples per
                collaborator. Defaults to 10.
            seed (int): Random numbers generator seed. Defaults to 0.
                For different splits on envoys, try setting different values
                for this parameter on each shard descriptor.
        """
        self.alpha = alpha
        self.min_samples_per_col = min_samples_per_col
        self.seed = seed

    def split(self, data, num_collaborators):
        """Split the data."""
        np.random.seed(self.seed)
        classes = len(np.unique(data))
        min_size = 0

        n = len(data)
        while min_size < self.min_samples_per_col:
            idx_batch = [[] for _ in range(num_collaborators)]
            for k in range(classes):
                idx_k = np.where(data == k)[0]
                np.random.shuffle(idx_k)
                proportions = np.random.dirichlet(np.repeat(self.alpha, num_collaborators))
                proportions = [
                    p * (len(idx_j) < n / num_collaborators)
                    for p, idx_j in zip(proportions, idx_batch)
                ]
                proportions = np.array(proportions)
                proportions = proportions / proportions.sum()
                proportions = (np.cumsum(proportions) * len(idx_k)).astype(int)[:-1]
                idx_splitted = np.split(idx_k, proportions)
                idx_batch = [idx_j + idx.tolist() for idx_j, idx in zip(idx_batch, idx_splitted)]
                min_size = min(len(idx_j) for idx_j in idx_batch)
        return idx_batch



# Copyright 2025 Intel Corporation
# SPDX-License-Identifier: Apache-2.0

"""This package contains utilities for PyTorch verifying datasets."""

from openfl.federated.data.sources.torch.verifiable_map_style_dataset import (
    VerifiableMapStyleDataset,
)
from openfl.federated.data.sources.torch.verifiable_map_style_image_folder import (
    VerifiableImageFolder,
)



# Copyright 2020-2024 Intel Corporation
# SPDX-License-Identifier: Apache-2.0


"""Custom input types definition for Click"""

import ast

import click

from openfl.utilities import utils


class FqdnParamType(click.ParamType):
    """Domain Type for click arguments.

    This class is used to validate that a command line argument is a fully
    qualified domain name.

    Attributes:
        name (str): The name of the parameter type.
    """

    name = "fqdn"

    def convert(self, value, param, ctx):
        """Validate value, if value is valid, return it.

        Args:
            value (str): The value to validate.
            param (click.core.Option): The option that this value was supplied
                to.
            ctx (click.core.Context): The context for the parameter.

        Returns:
            str: The value, if it is valid.

        Raises:
            value (click.exceptions.BadParameter): If the value is not a valid
                domain name.
        """
        if not utils.is_fqdn(value):
            self.fail(f"{value} is not a valid domain name", param, ctx)
        return value


class IpAddressParamType(click.ParamType):
    """IpAddress Type for click arguments.

    This class is used to validate that a command line argument is an IP
    address.

    Attributes:
        name (str): The name of the parameter type.
    """

    name = "IpAddress type"

    def convert(self, value, param, ctx):
        """Validate value, if value is valid, return it.

        Args:
            value (str): The value to validate.
            param (click.core.Option): The option that this value was supplied
                to.
            ctx (click.core.Context): The context for the parameter.

        Returns:
            str: The value, if it is valid.

        Raises:
            click.exceptions.BadParameter: If the value is not a valid IP
                address.
        """
        if not utils.is_api_adress(value):
            self.fail(f"{value} is not a valid ip address name", param, ctx)
        return value


class InputSpec(click.Option):
    """List or dictionary that corresponds to the input shape for a model"""

    def type_cast_value(self, ctx, value):
        try:
            if value is None:
                return None
            else:
                return ast.literal_eval(value)
        except Exception:
            raise click.BadParameter(value)


FQDN = FqdnParamType()
IP_ADDRESS = IpAddressParamType()



# Copyright (C) 2020-2021 Intel Corporation
# SPDX-License-Identifier: Apache-2.0
"""You may copy this file as the starting point of your own model."""

import logging
import os

import numpy as np

from .nii_reader import nii_reader

logger = logging.getLogger(__name__)


def train_val_split(features, labels, percent_train, shuffle):
    """Train/validation splot of the BraTS dataset.

    Splits incoming feature and labels into training and validation. The value
    of shuffle determines whether shuffling occurs before the split is performed.

    Args:
        features: The input images
        labels: The ground truth labels
        percent_train (float): The percentage of the dataset that is training.
        shuffle (bool): True = shuffle the dataset before the split

    Returns:
        train_features: The input images for the training dataset
        train_labels: The ground truth labels for the training dataset
        val_features: The input images for the validation dataset
        val_labels: The ground truth labels for the validation dataset
    """

    def split(lst, idx):
        """Split a Python list into 2 lists.

        Args:
            lst: The Python list to split
            idx: The index where to split the list into 2 parts

        Returns:
            Two lists

        """
        if idx < 0 or idx > len(lst):
            raise ValueError('split was out of expected range.')
        return lst[:idx], lst[idx:]

    nb_features = len(features)
    nb_labels = len(labels)
    if nb_features != nb_labels:
        raise RuntimeError('Number of features and labels do not match.')
    if shuffle:
        new_order = np.random.permutation(np.arange(nb_features))
        features = features[new_order]
        labels = labels[new_order]
    split_idx = int(percent_train * nb_features)
    train_features, val_features = split(lst=features, idx=split_idx)
    train_labels, val_labels = split(lst=labels, idx=split_idx)
    return train_features, train_labels, val_features, val_labels


def load_from_nifti(parent_dir,
                    percent_train,
                    shuffle,
                    channels_last=True,
                    task='whole_tumor',
                    **kwargs):
    """Load the BraTS dataset from the NiFTI file format.

    Loads data from the parent directory (NIfTI files for whole brains are
    assumed to be contained in subdirectories of the parent directory).
    Performs a split of the data into training and validation, and the value
    of shuffle determined whether shuffling is performed before this split
    occurs - both split and shuffle are done in a way to
    keep whole brains intact. The kwargs are passed to nii_reader.

    Args:
        parent_dir: The parent directory for the BraTS data
        percent_train (float): The percentage of the data to make the training dataset
        shuffle (bool): True means shuffle the dataset order before the split
        channels_last (bool): Input tensor uses channels as last dimension (Default is True)
        task: Prediction task (Default is 'whole_tumor' prediction)
        **kwargs: Variable arguments to pass to the function

    Returns:
        train_features: The input images for the training dataset
        train_labels: The ground truth labels for the training dataset
        val_features: The input images for the validation dataset
        val_labels: The ground truth labels for the validation dataset

    """
    path = os.path.join(parent_dir)
    subdirs = os.listdir(path)
    subdirs.sort()
    if not subdirs:
        raise SystemError(f'''{parent_dir} does not contain subdirectories.
Please make sure you have BraTS dataset downloaded
and located in data directory for this collaborator.
        ''')
    subdir_paths = [os.path.join(path, subdir) for subdir in subdirs]

    imgs_all = []
    msks_all = []
    for brain_path in subdir_paths:
        these_imgs, these_msks = nii_reader(
            brain_path=brain_path,
            task=task,
            channels_last=channels_last,
            **kwargs
        )
        # the needed files where not present if a tuple of None is returned
        if these_imgs is None:
            logger.debug(f'Brain subdirectory: {brain_path} did not contain the needed files.')
        else:
            imgs_all.append(these_imgs)
            msks_all.append(these_msks)

    # converting to arrays to allow for numpy indexing used during split
    imgs_all = np.array(imgs_all)
    msks_all = np.array(msks_all)

    # note here that each is a list of 155 slices per brain, and so the
    # split keeps brains intact
    imgs_all_train, msks_all_train, imgs_all_val, msks_all_val = train_val_split(
        features=imgs_all,
        labels=msks_all,
        percent_train=percent_train,
        shuffle=shuffle
    )
    # now concatenate the lists
    imgs_train = np.concatenate(imgs_all_train, axis=0)
    msks_train = np.concatenate(msks_all_train, axis=0)
    imgs_val = np.concatenate(imgs_all_val, axis=0)
    msks_val = np.concatenate(msks_all_val, axis=0)

    return imgs_train, msks_train, imgs_val, msks_val



# Copyright 2025 Intel Corporation
# SPDX-License-Identifier: Apache-2.0

from abc import abstractmethod

import numpy as np
import torch

from openfl.federated.data.sources.verifiable_dataset_info import (
    VerifiableDatasetInfo,
)


class VerifiableMapStyleDataset(torch.utils.data.Dataset):
    """Base class for different types of map style datasets."""

    def __init__(self, vds: VerifiableDatasetInfo, transform=None, verify_dataset_items=False):
        self.transform = transform
        self.verifiable_dataset_info = vds
        self.verify_dataset_items = verify_dataset_items
        self.datasets = self.create_datasets()

        # create indices for fast lookup
        lengths = list(map(len, self.datasets))
        self.cumulative_sizes = np.cumsum(lengths)

    def __getitem__(self, idx):
        # find which sub-dataset this index belongs to
        dataset_idx = self.cumulative_sizes.searchsorted(idx, side="right")
        # find the data in that sub-dataset
        data_idx = idx - self.cumulative_sizes[dataset_idx - 1] if dataset_idx > 0 else idx
        item = self.datasets[dataset_idx][data_idx]
        data_path = item["path"]

        if self.verify_dataset_items:
            item_hash = self.verifiable_dataset_info.data_sources[dataset_idx].compute_file_hash(
                data_path
            )
            if not self.verifiable_dataset_info.verify_single_file(data_path, item_hash):
                raise ValueError(f"Data integrity check failed for {data_path}")

        return item["data"], item["label"]

    def __len__(self):
        return self.cumulative_sizes[-1] if len(self.cumulative_sizes) > 0 else 0

    @abstractmethod
    def create_datasets(self):
        """
        Create datasets from the data sources.

        This method should be implemented by subclasses to generate a list of datasets
        based on the provided data sources in the `VerifiableDatasetInfo` object.

        Returns:
            list: A list of datasets, where each dataset is expected to be an iterable
                  containing items in the form of dictionaries with keys "data", "label", "path".
        """
        pass



# Copyright 2020-2024 Intel Corporation
# SPDX-License-Identifier: Apache-2.0


"""Federated averaging module."""

import numpy as np

from openfl.interface.aggregation_functions.core import AggregationFunction


def weighted_average(tensors, weights):
    """Compute average."""
    return np.average(tensors, weights=weights, axis=0)


class WeightedAverage(AggregationFunction):
    """Weighted average aggregation."""

    def call(self, local_tensors, *_) -> np.ndarray:
        """Aggregate tensors.

        Args:
            local_tensors (list[openfl.utilities.LocalTensor]): List of local
                tensors to aggregate.
            db_iterator: iterator over history of all tensors. Columns:
                - 'tensor_name': name of the tensor.
                    Examples for `torch.nn.Module`s: 'conv1.weight','fc2.bias'.
                - 'round': 0-based number of round corresponding to this
                    tensor.
                - 'tags': tuple of tensor tags. Tags that can appear:
                    - 'model' indicates that the tensor is a model parameter.
                    - 'trained' indicates that tensor is a part of a training
                        result.
                        These tensors are passed to the aggregator node after
                        local learning.
                    - 'aggregated' indicates that tensor is a result of
                        aggregation.
                        These tensors are sent to collaborators for the next
                        round.
                    - 'delta' indicates that value is a difference between
                        rounds for a specific tensor.
                    also one of the tags is a collaborator name
                    if it corresponds to a result of a local task.

                - 'nparray': value of the tensor.
            tensor_name: name of the tensor
            fl_round: round number
            tags: tuple of tags for this tensor
        Returns:
            np.ndarray: aggregated tensor
        """
        tensors = np.asarray([x.tensor for x in local_tensors], dtype=np.float32)
        weights = np.asarray([x.weight for x in local_tensors], dtype=np.float32)
        return weighted_average(tensors, weights)



# Copyright 2025 Intel Corporation
# SPDX-License-Identifier: Apache-2.0
"""This package contains utilities for verifying datasets."""

from openfl.federated.data.sources.data_source import DataSource, DataSourceType
from openfl.federated.data.sources.local_data_source import LocalDataSource
from openfl.federated.data.sources.s3_data_source import S3DataSource
from openfl.federated.data.sources.verifiable_dataset_info import VerifiableDatasetInfo



# Copyright 2025 Intel Corporation
# SPDX-License-Identifier: Apache-2.0

"""Base class for different types of data sources."""

import hashlib
from abc import ABC, abstractmethod
from enum import Enum, auto
from pathlib import Path
from typing import Any, Dict, Generator


class DataSourceType(Enum):
    """Enum for the different types of data sources."""

    LOCAL = auto()
    S3 = auto()


class DataSource(ABC):
    """
    Base class for different types of data sources.

    Attributes:
        type (str): The storage type of the data source
    """

    def __init__(self, type: DataSourceType):
        """
        Initialize a DataSource.

        Args:
            type (DataSourceType): The storage type of the data source.
        """
        self.type = type

    @abstractmethod
    def compute_file_hash(self, path: str) -> str:
        """
        Compute the hash of the object or file.

        Args:
            path (str): Path to the file.

        Returns:
            str: The file's hash.
        """
        pass

    @abstractmethod
    def enumerate_files(self) -> Generator[str, None, None]:
        """
        Enumerate all files in the data source.

        Yields:
            str: The file path.
        """
        pass

    @abstractmethod
    def read_blob(self, path: str) -> bytes:
        """
        Read a blob from the data source.
        Args:
            path (str): Path to the file.
        Returns:
            bytes: The file's content.
        """
        pass

    @classmethod
    @abstractmethod
    def from_dict(cls, ds_dict: dict):
        """
        Create a DataSource from a dictionary.

        Args:
            ds_dict (dict): The dictionary to convert.

        Returns:
            DataSource: The created DataSource.
        """
        pass

    def is_valid_hash_function(self, func):
        try:
            func_name = func.__name__.removeprefix("openssl_")

            # Ensure it is a known hashlib function
            if func_name not in hashlib.algorithms_available:
                return False

            # Ensure it's callable and returns a valid hash object
            test_instance = func()
            return callable(func) and hasattr(test_instance, "digest")

        except Exception:
            return False

    def _serialize_fields(self) -> Dict[str, Any]:
        """Returns a dictionary of serializable fields."""
        serializable_dict = {}
        for key, val in self.__dict__.items():
            if key.startswith("_"):  # Skip private attributes
                continue
            if self.is_valid_hash_function(val):
                val = val.__name__.removeprefix("openssl_")
            if callable(val):
                continue  # Skip methods
            if isinstance(val, Path):
                val = str(val)  # Convert Path to string
            elif isinstance(val, Enum):
                val = val.value  # Convert Enum to its value
            serializable_dict[key] = val
        return serializable_dict

    def to_dict(self) -> Dict[str, Any]:
        """Convert the object to a dictionary using the serialization rules."""
        return self._serialize_fields()



# Copyright 2020-2024 Intel Corporation
# SPDX-License-Identifier: Apache-2.0


"""DataLoader module."""


class DataLoader:
    """A base class used to represent a Federated Learning Data Loader.

    This class should be inherited by any data loader class specific to a
    machine learning framework.

    Attributes:
        None
    """

    def __init__(self, **kwargs):
        """Initializes the DataLoader object.

        Args:
            kwargs: Additional arguments to pass to the function.
        """
        pass

    def get_feature_shape(self):
        """Returns the shape of an example feature array.

        Raises:
            NotImplementedError: This method must be implemented by a child
                class.
        """
        raise NotImplementedError

    def get_train_loader(self, **kwargs):
        """Returns the data loader for the training data.

        Args:
            kwargs: Additional arguments to pass to the function.

        Raises:
            NotImplementedError: This method must be implemented by a child
                class.
        """
        raise NotImplementedError

    def get_valid_loader(self):
        """Returns the data loader for the validation data.

        Raises:
            NotImplementedError: This method must be implemented by a child
                class.
        """
        raise NotImplementedError

    def get_infer_loader(self):
        """Returns the data loader for inferencing data.

        Raises:
            NotImplementedError: This method must be implemented by a child
                class.
        """
        return NotImplementedError

    def get_train_data_size(self):
        """Returns the total number of training samples.

        Raises:
            NotImplementedError: This method must be implemented by a child
                class.
        """
        raise NotImplementedError

    def get_valid_data_size(self):
        """Returns the total number of validation samples.

        Raises:
            NotImplementedError: This method must be implemented by a child
                class.
        """
        raise NotImplementedError



# Copyright 2020-2024 Intel Corporation
# SPDX-License-Identifier: Apache-2.0


"""RandomShiftPipeline module."""

import numpy as np

from openfl.pipelines.pipeline import Float32NumpyArrayToBytes, TransformationPipeline, Transformer


class RandomShiftTransformer(Transformer):
    """Random Shift Transformer class."""

    def __init__(self):
        """Initialize RandomShiftTransformer."""
        self.lossy = False

        return

    def forward(self, data, **kwargs):
        """Forward pass - compress data.

        Implement the data transformation.

        Args:
            data: The data to be transformed.

        Returns:
            transformed_data: The data after the random shift.
            metadata: The metadata for the transformation.
        """
        shape = data.shape
        random_shift = np.random.uniform(low=-20, high=20, size=shape).astype(np.float32)
        transformed_data = data + random_shift

        # construct metadata
        metadata = {"int_to_float": {}, "int_list": list(shape)}
        for idx, val in enumerate(random_shift.flatten(order="C")):
            metadata["int_to_float"][idx] = val

        return transformed_data, metadata

    def backward(self, data, metadata, **kwargs):
        """Backward pass - Decompress data.

        Implement the data transformation needed when going the oppposite
        direction to the forward method.

        Args:
            data: The transformed data.
            metadata: The metadata for the transformation.

        Returns:
            The original data before the random shift.
        """
        shape = tuple(metadata["int_list"])
        # this is an awkward use of the metadata into to float dict, usually
        # it will truly be treated as a dict. Here (and in 'forward' above)
        # we use it essentially as an array.
        shift = np.reshape(
            np.array(
                [metadata["int_to_float"][idx] for idx in range(len(metadata["int_to_float"]))]
            ),
            newshape=shape,
            order="C",
        )
        return data - shift


class RandomShiftPipeline(TransformationPipeline):
    """Random Shift Pipeline."""

    def __init__(self, **kwargs):
        """Initialize."""
        transformers = [RandomShiftTransformer(), Float32NumpyArrayToBytes()]
        super().__init__(transformers=transformers)



# Copyright 2020-2024 Intel Corporation
# SPDX-License-Identifier: Apache-2.0


"""Adam optimizer module."""

from typing import Dict, Optional, Tuple

import numpy as np

from openfl.utilities.optimizers.numpy.base_optimizer import Optimizer


class NumPyAdam(Optimizer):
    """Adam optimizer implementation.

    Implements the Adam optimization algorithm using NumPy.
    Adam is an algorithm for first-order gradient-based optimization of
    stochastic objective functions, based on adaptive estimates of lower-order
    moments.

    Original paper: https://openreview.net/forum?id=ryQu7f-RZ

    Attributes:
        params (dict, optional): Parameters to be stored for optimization.
        model_interface: Model interface instance to provide parameters.
        learning_rate (float): Tuning parameter that determines the step size
            at each iteration.
        betas (tuple): Coefficients used for computing running averages of
            gradient and its square.
        initial_accumulator_value (float): Initial value for gradients and
            squared gradients.
        epsilon (float): Value for computational stability.
    """

    def __init__(
        self,
        *,
        params: Optional[Dict[str, np.ndarray]] = None,
        model_interface=None,
        learning_rate: float = 0.01,
        betas: Tuple[float, float] = (0.9, 0.999),
        initial_accumulator_value: float = 0.0,
        epsilon: float = 1e-8,
    ) -> None:
        """Initialize the Adam optimizer.

        Args:
            params (dict, optional): Parameters to be stored for optimization.
                Defaults to None.
            model_interface: Model interface instance to provide parameters.
                Defaults to None.
            learning_rate (float, optional): Tuning parameter that determines
                the step size at each iteration. Defaults to 0.01.
            betas (tuple, optional): Coefficients used for computing running
                averages of gradient and its square. Defaults to (0.9, 0.999).
            initial_accumulator_value (float, optional): Initial value for
                gradients and squared gradients. Defaults to 0.0.
            epsilon (float, optional): Value for computational stability.
                Defaults to 1e-8.

        Raises:
            ValueError: If both params and model_interface are None.
            ValueError: If learning_rate is less than 0.
            ValueError: If betas[0] is not in [0, 1).
            ValueError: If betas[1] is not in [0, 1).
            ValueError: If initial_accumulator_value is less than 0.
            ValueError: If epsilon is less than or equal to 0.
        """
        super().__init__()

        if model_interface is None and params is None:
            raise ValueError("Should provide one of the params or model_interface")

        if learning_rate < 0:
            raise ValueError(f"Invalid learning rate: {learning_rate}. Learning rate must be >= 0.")
        if not 0.0 <= betas[0] < 1:
            raise ValueError(f"Invalid betas[0] value: {betas[0]}. betas[0] must be in [0, 1).")
        if not 0.0 <= betas[1] < 1:
            raise ValueError(f"Invalid betas[1] value: {betas[1]}. betas[1] must be in [0, 1).")
        if initial_accumulator_value < 0:
            raise ValueError(
                f"Invalid initial_accumulator_value value: {initial_accumulator_value}. \
                Initial accumulator value must be >= 0."
            )
        if epsilon <= 0:
            raise ValueError(f"Invalid epsilon value: {epsilon}. Epsilon avalue must be > 0.")

        self.params = params

        if params is None and model_interface is not None:
            self._set_params_from_model(model_interface)

        self.learning_rate = learning_rate
        self.beta_1, self.beta_2 = betas
        self.initial_accumulator_value = initial_accumulator_value
        self.epsilon = epsilon
        self.current_step = dict.fromkeys(self.params, 0)

        self.grads_first_moment, self.grads_second_moment = {}, {}

        for param_name in self.params:
            self.grads_first_moment[param_name] = np.full_like(
                self.params[param_name], self.initial_accumulator_value
            )
            self.grads_second_moment[param_name] = np.full_like(
                self.params[param_name], self.initial_accumulator_value
            )

    def _update_first_moment(self, grad_name: str, grad: np.ndarray) -> None:
        """
        Update gradients first moment.

        Args:
            grad_name (str): The name of the gradient.
            grad (np.ndarray): The gradient values.
        """
        self.grads_first_moment[grad_name] = self.beta_1 * self.grads_first_moment[grad_name] + (
            (1.0 - self.beta_1) * grad
        )

    def _update_second_moment(self, grad_name: str, grad: np.ndarray) -> None:
        """
        Update gradients second moment.

        Args:
            grad_name (str): The name of the gradient.
            grad (np.ndarray): The gradient values.
        """
        self.grads_second_moment[grad_name] = self.beta_2 * self.grads_second_moment[grad_name] + (
            (1.0 - self.beta_2) * grad**2
        )

    def step(self, gradients: Dict[str, np.ndarray]) -> None:
        """Perform a single step for parameter update.

        Implement Adam optimizer weights update rule.

        Args:
            gradients (dict): Partial derivatives with respect to optimized
                parameters.

        Raises:
            KeyError: If a key in gradients does not exist in optimized
                parameters.
        """
        for grad_name in gradients:
            if grad_name not in self.grads_first_moment:
                raise KeyError(f"Key {grad_name} doesn't exist in optimized parameters")

            grad = gradients[grad_name]

            self._update_first_moment(grad_name, grad)
            self._update_second_moment(grad_name, grad)

            t = self.current_step[grad_name] + 1
            mean = self.grads_first_moment[grad_name]
            var = self.grads_second_moment[grad_name]

            grads_first_moment_normalized = mean / (1.0 - self.beta_1**t)
            grads_second_moment_normalized = var / (1.0 - self.beta_2**t)

            # Make an update for a group of parameters
            self.params[grad_name] -= (
                self.learning_rate
                * grads_first_moment_normalized
                / (np.sqrt(grads_second_moment_normalized) + self.epsilon)
            )

            self.current_step[grad_name] += 1



# Copyright 2020-2024 Intel Corporation
# SPDX-License-Identifier: Apache-2.0


"""openfl.experimental.workflow.interface.participants module."""

from typing import Any, Callable, Dict, Optional


class Participant:
    """Class for a participant.

    Attributes:
        private_attributes (dict): The private attributes of the participant.
        _name (str): The name of the participant.
    """

    def __init__(self, name: str = ""):
        """Initializes the Participant object with an optional name.

        Args:
            name (str, optional): The name of the participant. Defaults to "".
        """
        self.private_attributes = {}
        self._name = name

    @property
    def name(self):
        """Returns the name of the participant.

        Returns:
            str: The name of the participant.
        """
        return self._name

    @name.setter
    def name(self, name: str):
        """Sets the name of the participant.

        Args:
            name (str): The name to be set.
        """
        self._name = name

    def get_name(self) -> str:
        """Gets the name of Participant (aggregator or collaborator).

        Returns:
            str: The name of the aggregator or collaborator.
        """
        return self._name

    def initialize_private_attributes(self, private_attrs: Dict[Any, Any] = None) -> None:
        """Initialize private attributes of Participant (aggregator or collaborator)
        by invoking the callable specified by user."""
        if self.private_attributes_callable is not None:
            self.private_attributes = self.private_attributes_callable(**self.kwargs)
        elif private_attrs:
            self.private_attributes = private_attrs

    def _set_private_attrs_to_clone(self, clone: Any) -> None:
        """Set private attributes to FLSpec clone before
        transitioning from aggregator to collaborator or vice versa

        Args:
            clone (Any): The clone to set attributes to.
        """
        # set private attributes as clone attributes
        for name, attr in self.private_attributes.items():
            setattr(clone, name, attr)

    def _delete_private_attrs_from_clone(self, clone: Any) -> None:
        """Remove private attributes from FLSpec clone before
        transitioning from collaborator to aggregator or vice versa

        Args:
            clone (Any): The clone to remove attributes from.
        """
        # Update  private attributes by taking latest
        # parameters from clone, then delete attributes from clone.
        for attr_name in self.private_attributes:
            if hasattr(clone, attr_name):
                self.private_attributes.update({attr_name: getattr(clone, attr_name)})
                delattr(clone, attr_name)

    def private_attributes(self, attrs: Dict[str, Any]) -> None:
        """Set the private attributes of the participant. These attributes will
        only be available within the tasks performed by the participants and
        will be filtered out prior to the task's state being transferred.

        Args:
            attrs (Dict[str, Any]): dictionary of ATTRIBUTE_NAME (str) ->
                object that will be accessible within the participant's task.

                Example:
                {'train_loader' : torch.utils.data.DataLoader(...)}

                In any task performed by this participant performed within the
                flow, this attribute could be referenced with self.train_loader
        """
        self.private_attributes = attrs


class Collaborator(Participant):
    """Class for a collaborator participant, derived from the Participant
    class.

    Attributes:
        name (str): Name of the collaborator.
        private_attributes_callable (Callable): A function which returns
            collaborator private attributes for each collaborator.
        num_cpus (int): Specifies how many cores to use for the collaborator
            step execution.
        num_gpus (float): Specifies how many GPUs to use to accelerate the
            collaborator step execution.
        kwargs (dict): Parameters required to call private_attributes_callable
            function.
    """

    def __init__(
        self,
        name: str = "",
        private_attributes_callable: Callable = None,
        num_cpus: int = 0,
        num_gpus: int = 0.0,
        **kwargs,
    ):
        """Initializes the Collaborator object.

        Create collaborator object with custom resources and a callable
        function to assign private attributes.

        Args:
            name (str, optional): Name of the collaborator. Defaults to "".
            private_attributes_callable (Callable, optional): A function which
                returns collaborator private attributes for each collaborator.
                In case private_attributes are not required this can be
                omitted. Defaults to None.
            num_cpus (int, optional): Specifies how many cores to use for the
                collaborator step execution. This will only be used if backend
                is set to ray. Defaults to 0.
            num_gpus (float, optional): Specifies how many GPUs to use to
                accelerate the collaborator step execution. This will only be
                used if backend is set to ray. Defaults to 0.0.
            **kwargs (dict): Parameters required to call
                private_attributes_callable function. The key of the
                dictionary must match the arguments to the
                private_attributes_callable. Defaults to {}.
        """
        super().__init__(name=name)
        self.num_cpus = num_cpus
        self.num_gpus = num_gpus
        self.kwargs = kwargs

        if private_attributes_callable is None:
            self.private_attributes_callable = private_attributes_callable
        else:
            if not callable(private_attributes_callable):
                raise Exception("private_attributes_callable  parameter must be a callable")
            else:
                self.private_attributes_callable = private_attributes_callable

    def execute_func(self, ctx: Any, f_name: str, callback: Callable) -> Any:
        """Execute remote function f.

        Args:
            ctx (Any): The context to execute the function in.
            f_name (str): The name of the function to execute.
            callback (Callable): The callback to execute after the function.

        Returns:
            Any: The result of the function execution.
        """
        self._set_private_attrs_to_clone(ctx)

        callback(ctx, f_name)

        self._delete_private_attrs_from_clone(ctx)

        return ctx


class Aggregator(Participant):
    """Class for an aggregator participant, derived from the Participant
    class."""

    def __init__(
        self,
        name: str = "",
        private_attributes_callable: Callable = None,
        num_cpus: int = 0,
        num_gpus: int = 0.0,
        **kwargs,
    ):
        """Initializes the Aggregator object.

        Create aggregator object with custom resources and a callable
        function to assign private attributes.

        Args:
            name (str, optional): Name of the aggregator. Defaults to "".
            private_attributes_callable (Callable, optional): A function which
                returns aggregator private attributes. In case
                private_attributes are not required this can be omitted.
                Defaults to None.
            num_cpus (int, optional): Specifies how many cores to use for the
                aggregator step execution. This will only be used if backend
                is set to ray. Defaults to 0.
            num_gpus (float, optional): Specifies how many GPUs to use to
                accelerate the aggregator step execution. This will only be
                used if backend is set to ray. Defaults to 0.0.
            **kwargs: Parameters required to call private_attributes_callable
                function. The key of the dictionary must match the arguments
                to the private_attributes_callable. Defaults to {}.
        """
        super().__init__(name=name)
        self.num_cpus = num_cpus
        self.num_gpus = num_gpus
        self.kwargs = kwargs

        if private_attributes_callable is None:
            self.private_attributes_callable = private_attributes_callable
        else:
            if not callable(private_attributes_callable):
                raise Exception("private_attributes_callable parameter must be a callable")
            else:
                self.private_attributes_callable = private_attributes_callable

    def execute_func(
        self, ctx: Any, f_name: str, callback: Callable, clones: Optional[Any] = None
    ) -> Any:
        """Executes remote function f.

        Args:
            ctx (Any): The context to execute the function in.
            f_name (str): The name of the function to execute.
            callback (Callable): The callback to execute after the function.
            clones (Optional[Any], optional): The clones to use in the
                function. Defaults to None.

        Returns:
            Any: The result of the function execution.
        """
        self._set_private_attrs_to_clone(ctx)

        if clones is not None:
            callback(ctx, f_name, clones)
        else:
            callback(ctx, f_name)

        self._delete_private_attrs_from_clone(ctx)

        return ctx
